#!/usr/bin/python
import os, sys, cPickle, gzip, time, stat

cdcat = []

def usage():
     print 'CD/DVD catalog: maintain a small file database (name,size,time)'
     print 'usage: pycdcat <command> <dbname> [...]'
     print 'commands: a[dd] l[ist] d[el]'
     print 'pycdcat a <dbname> <dir> <volname>'
     print 'pycdcat l <dbname> [<columnorder>]'
     print 'pycdcat d <dbname> <volname>'
     print 'examples:\npycdcat a mycdromdb /mnt/cdrom Backup_July'
     print 'pycdcat l mycdromdb snt'
     sys.exit(0)

def savedb(dbname):
    dbfile = gzip.open(dbname, 'wb')
    cPickle.dump(cdcat, dbfile, 0)
    dbfile.close()

def appenddir(dir, volname, mountdir):
    for f in os.listdir(dir):
        pathname = '%s%s%s' % (dir, os.sep, f)
        mode = os.lstat(pathname)[stat.ST_MODE]
        if stat.S_ISDIR(mode): appenddir(pathname, volname, mountdir)
        elif stat.S_ISREG(mode):
            cdcat.append((pathname.replace(dirname, volname, 1),
                os.lstat(pathname)[stat.ST_SIZE],
                os.lstat(pathname)[stat.ST_MTIME] ))

def delvolume(volume):
    #if len(cdcat) < 1: return
    volume += os.sep
    delrange = [ ]
    for i,f in enumerate(cdcat):
        if f[0].startswith(volume):
            delrange.append(i)
    if len(delrange) > 0:
        del cdcat[min(delrange):max(delrange)+1]

if __name__ == '__main__':
    try:
        cmd = sys.argv[1]
        dbname = sys.argv[2]
    except:
        usage()
    try:
        dbfile = gzip.open(dbname, 'rb')
        cdcat = cPickle.load(dbfile)
        dbfile.close()
    except:
        print "WARNING: couldn't read file, start with empty db"
    if cmd == 'a' or cmd == 'add':
        dirname = sys.argv[3].rstrip(os.sep)
        volname = sys.argv[4].rstrip(os.sep)
        delvolume(volname)
        appenddir(dirname, volname, dirname)
        savedb(dbname)
    elif cmd == 'l' or cmd == 'list':
        try:
            order = sys.argv[3]
        except:
            order = 'nst'
        tupleorder = { 'n': 0, 's': 1, 't': 2 }  #=dict(n=0, s=1, t=2) #for python > 2.2
        def colstring(ft, col):
            if col != 't': return '%s' % ft[tupleorder[col]]
            return '%s' % time.strftime("%Y%m%d%H%M",time.localtime(ft[tupleorder[col]]))
            # '%s' % time.ctime(ft[tupleorder[col]])
        for f in cdcat:
            #print ' '.join(['%s' % (f[tupleorder[o]]) for o in order])
            print ' '.join([colstring(f,o) for o in order])
    elif cmd == 'd' or cmd == 'del' or cmd == 'delete':
        volname = sys.argv[3].rstrip(os.sep)
        delvolume(volname)
        savedb(dbname)
    elif cmd == 'lv':
        vols = []
        vol2 = ''
        count = 0
        for f in cdcat:
            vol = f[0][:f[0].index('/')]
            if vol not in vols:
                if vol2 != '': print '%d %s' % (count, vol2)
                count = 1
                vols.append(vol)
                vol2 = vol
            else: count += 1
        print '%d %s' % (count, vol2)
    elif cmd == 'rv' or cmd == 'renamevolume':
        oldname = sys.argv[3].rstrip(os.sep)
        oldname += '/'
        newname = sys.argv[4].rstrip(os.sep)
        newname += '/'
        for i,f in enumerate(cdcat):
            if f[0].startswith(oldname):
                fname = f[0]
                fname = fname.replace(oldname, newname, 1)
                cdcat[i] = (fname, f[1], f[2])
        savedb(dbname)
    elif cmd == 'm' or cmd == 'merge':
        mergename = sys.argv[3]
        mergefile = gzip.open(mergename, 'rb')
        mergecat = cPickle.load(mergefile)
        mergefile.close()
        vol2 = ''
        for f in mergecat:
            vol = f[0][:f[0].index('/')]
            if vol != vol2: delvolume(vol)
            vol2 = vol
            cdcat.append(f)
        savedb(dbname)
