#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
/*
 * FW Bergmann <bergmann@wasenmeister.de>
 * cdr2wav.c, v. 0.0.1
 * fast .cdr to .wav file conversion
 */

#define BLOCKLEN 2352
#define BUFLEN 32768
static const unsigned char header[] = {
  'R', 'I', 'F', 'F', 0x54, 0x9, 0x0, 0x0, 'W', 'A', 'V', 'E', 'f', 'm', 't', ' ',
  0x10, 0x0, 0x0, 0x0, 0x1, 0x0, 0x2, 0x0, 0x44, 0xac, 0x0, 0x0, 0x10, 0xb1, 0x2, 0x0,
  0x4, 0x0, 0x10, 0x0, 'd', 'a', 't', 'a', 0x30, 0x9, 0x0, 0x0 };
static const char usg[] =
  "usage:\n"
  "cdr2wav input_audio.cdr output_audio.wav [startframe [countframes]]\n";

int str_len(const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (int)(t-in);
}

void usage() {
  write(1, usg, str_len(usg));
  exit(1);
}

/* swap_bytes
 * actually we don't swap bytes but change the order, from
 *    xxxx MSB1 LSB1 MSB2 LSB2 ...    to
 *    LSB1 MSB1 LSB2 MSB2 LSB3 ...
 */
void swap_bytes(char *buf, int bytes) {
  register char *p;
  register char *q;
  p = buf;
  q = p+bytes;
  for(;;) {
    *p = p[2]; p += 2; if (p>=q) break;
    *p = p[2]; p += 2; if (p>=q) break;
    *p = p[2]; p += 2; if (p>=q) break;
    *p = p[2]; p += 2; if (p>=q) break;
    /* b = *p; *p = p[1]; ++p; *p = b; ++p; if (p>=q) break; */
  }
}

ssize_t convert(int ifd,int ofd) {
  register ssize_t bytes, total;
  register char *c, *w;
  register int i,o;
  char audiobuf[BUFLEN+2];
  w = audiobuf;
  c = w+1;
  total = 0;
  i = ifd;
  o = ofd;
  for (;;) {
    bytes = read(i, c, BUFLEN);
    if (bytes<1) break;
    swap_bytes(w, bytes);
    write(o, w, bytes);
    total += bytes;
  }
  return total;
}

ssize_t convertn(int i, int o, ssize_t limit) {
  register char *p, *q;
  register ssize_t bytes;
  register ssize_t total=0;
  ssize_t l=limit;
  ssize_t toread=BUFLEN;
  char audiobuf[BUFLEN+2];

  p = audiobuf;
  for (;;) {
    if ((l-total)<toread) toread = l-total;
    q = p+1;
    bytes = read(i, q, toread);
    if (bytes<1) break;
    q = p+bytes;
    for(;;) {
      *p = p[2]; p += 2; if (p>=q) break;
      *p = p[2]; p += 2; if (p>=q) break;
      *p = p[2]; p += 2; if (p>=q) break;
      *p = p[2]; p += 2; if (p>=q) break;
    }
    p = audiobuf;
    write(o, p, bytes);
    total += bytes;
    if (total>=l) break;
  }
  return total;
}

int main(int argc, char* argv[]) {
  register int i,o;
  register char *bin;
  register char *bout;
  char audiobuf[BUFLEN+2];
  char *cdr, *wav;
  long fstart, fcount;
  int ret;
  ssize_t written;
  uint32_t len32;
  
  if (argc<3) usage();
  cdr = argv[1];
  wav = argv[2];
  bout = audiobuf;
  bin = bout+1;
  ret = 2;
  fstart = fcount = 0;
  if (argc>3) fstart = atol(argv[3]);
  if (argc>4) fcount = atol(argv[4]);
  fstart *= BLOCKLEN;
  fcount *= BLOCKLEN;
  if ((i=open(cdr, O_RDONLY))>0) {
    ret++;
    if (lseek(i, fstart, SEEK_SET)==fstart) {
      ret++;
      if ((o=open(wav, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH))>0) {
        write(o, header, 44);
        if (fcount) {
          written = convertn(i, o, fcount);
        } else {
          written = convert(i, o);
        }
        len32 = written;
        lseek(o, 40, SEEK_SET); write(o, &len32, 4);
        len32 += 36;
        lseek(o, 4, SEEK_SET); write(o, &len32, 4);
        close(o);
        ret = 0;
      }
    }
    close(i);
  }
  return ret;
}
