#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

/*
 * FW Bergmann <something@tuxad.com>
 * lgrep.c v. 0.0.5
 * Thanks to djb and fefe!
 */

#define LINE_LEN 1024
#define BUFFER_LEN 8192

typedef struct buffer {
  char *x;              /* start of buffer data */
  unsigned long p;      /* current position in buffer */
  unsigned long n;      /* current size of data in buffer */
  unsigned long a;      /* size of buffer data space */
  int fd;               /* file descriptor */
} buffer;

char buffer_1_space[BUFFER_LEN+1];
static buffer ot = { (buffer_1_space), 0, 0, (sizeof buffer_1_space), 1};
buffer *buffer_1 = &ot;
char buffer_2_space[BUFFER_LEN+1];
static buffer et = { (buffer_2_space), 0, 0, (sizeof buffer_2_space), 2};
buffer *buffer_2 = &et;

void byte_copy( char *to, unsigned long len, const char *from) {
  register const char *s=from;
  register char *t=to;
  register char *u=t+len;
  for (;;) {
    if (t>=u) break; *t=*s; ++t; ++s;
    if (t>=u) break; *t=*s; ++t; ++s;
    if (t>=u) break; *t=*s; ++t; ++s;
    if (t>=u) break; *t=*s; ++t; ++s;
  }
  return;
}

unsigned int strlen(const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (unsigned int)(t-in);
}

unsigned int linelen(const char* in) {
  register const char* t=in;
  for (;;) {
    if (*t=='\n') break; ++t;
    if (*t=='\n') break; ++t;
    if (*t=='\n') break; ++t;
    if (*t=='\n') break; ++t;
  }
  return (unsigned int)(t-in+1);
}

int str_diffn(const char* a, const char* b, unsigned long limit) {
  register const unsigned char* s=(const unsigned char*)a;
  register const unsigned char* t=(const unsigned char*)b;
  register const unsigned char* u=t+limit;
  register int j;
  j=0;
  for (;;) { if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t; }
  return j;
}

void buffer_init( buffer *b, char *x, unsigned long a, int fd) {
  b->fd=fd;
  b->x=x;
  b->a=a;
  b->n=0;
  b->p=0;
  return;
}

int buffer_feed( buffer *b) {
  register int w;
  for(;;) {
    if((w=read(b->fd, b->x, b->a))<0) {
      if(errno==EINTR) continue;
      else return w;
    }
    break;
  }
  b->n=w;
  b->p=0;
  return w;
}

int buffer_flush(buffer* b) {
  register char *buf;
  register unsigned long len;
  register int w;

  /* if (b->fd==1) printf("flush: p=%lu n=%lu a=%lu\n", b->p, b->n, b->a); */
  len=b->n;
  buf=b->x+b->p-len;
  while(len) {
    if((w=write(b->fd,buf,len))<0) {
      if(errno==EINTR) continue;
      return -1;
    }
    buf+=w;
    len-=w;
    b->n=len;
  }
  b->p=0;
  return 0;
}

/*
int buffer_getc(buffer *b, char *x) {
  if(b->p==b->n) {
    register int blen;
    if((blen=buffer_feed(b))<=0) return blen;
  }
  *x=b->x[b->p];
  ++b->p;
  return 1;
}
*/

int buffer_put( buffer *b, const char *buf, unsigned long len) {
  if (len>b->a-b->p) {
    if (buffer_flush(b)<0) return -1;
    if (len>b->a) {
      register int w;
      while (len) {
        if ((w=write(b->fd,buf,len))<0) {
          if (errno==EINTR) continue;
          return -1;
        }
        buf += w;
        len -= w;
      }
      return 0;
    }
  }
  byte_copy( b->x+b->p, len, buf);
  b->p += len;
  b->n += len;
  return 0;
}

int buffer_puts( buffer *b, const char *str) {
  return buffer_put( b, str, strlen(str));
}

int line_cmp_old(register const char *s1,register const char *s2,register ssize_t n) {
  register char c;

  if (!n) return 0;
  --n;
  for(;;) {
    c = *s1; if (!n) break; if (c != *s2) break; ++s1; ++s2; --n;
  }
  return ((int)(unsigned int)(unsigned char) c) - ((int)(unsigned int)(unsigned char) *s2);
}

ssize_t read_line(int fd,char *buf,register ssize_t n) {
  register ssize_t c,b;
  register int line;

  c = 0;
  line = 1;
  b = 1;
  for(;;) {
    b = read(fd,buf+c,1);
    if (b!=1) return b;
    if (buf[c] == '\n') if (c != 0) break;
    ++c; if (c >= n) break;
  }
  return c;
}

long line_array_match(
    const char **lines,register long lines_count,
    const char *line,unsigned long len) {
  register long i;

  i = 0;
  for(;;) {
    if (i >= lines_count) break; if (str_diffn(line,lines[i],len)==0) return i; ++i;
    if (i >= lines_count) break; if (str_diffn(line,lines[i],len)==0) return i; ++i;
    if (i >= lines_count) break; if (str_diffn(line,lines[i],len)==0) return i; ++i;
    if (i >= lines_count) break; if (str_diffn(line,lines[i],len)==0) return i; ++i;
  }
  return -1;
}

unsigned long line_count(char *buf,unsigned long len) {
  register char *cur;
  register char *buf_end;
  register unsigned long c;

  cur = buf;
  buf_end = buf + len;
  c = 0;
  for(;;) {
    if (cur >= buf_end) break; if (*cur == '\n') ++c; ++cur;
  }
  return c;
}

void line_array_fill(const char *buf,unsigned long len,register const char **lines) {
  register const char *cur;
  register const char *buf_end;
  register long unsigned i;

  cur = buf;
  buf_end = cur + len;
  lines[0] = cur;
  i = 1;
  for(;;) {
    if (cur >= buf_end) break;
    if (*cur == '\n') {
      ++cur;
      lines[i] = cur;
      ++i;
    } else {
      ++cur;
    }
  }
}

int buffer_getline( buffer *b, char *x, unsigned long len) {
  register unsigned long blen;
  register char c;

  for (blen=1;;++blen) {
    if (b->p==b->n) {
      register int r;
      r = buffer_feed(b);
      if (r==0) { *x=0; return blen-1; }
      if (r<0) return r;
    }
    c = b->x[b->p];
    ++b->p;
    if (blen<len) { *x = c; ++x; }
    if (c=='\n') break;
  }
  return blen;
}

void find_lines( register const char **f2lines,
    const unsigned long f2lcount, buffer *sbuf) {
  char line[LINE_LEN+1];
  register char *p;
  register unsigned long len;

  p = line;
  for(;;) {
    len = buffer_getline( sbuf, p, LINE_LEN);
    if (len<=0) break;
    if (line_array_match(f2lines,f2lcount,p,len)<0) {
      buffer_put( buffer_2, p, len);
    } else {
      buffer_put( buffer_1, p, len);
    }
  }
  return;
}

void usage() {
  buffer_puts(buffer_1, "usage:\n"
           "ipgrep linesfile [srcfile]\n"
           "writes lines from srcfile also contained in linesfile to stdout\n"
           "writes lines from srcfile not contained in linesfile to stderr\n"
           "if srcfile is not specified then stdin will be used\n");
  _exit(1);
}

void err_exit(const char *s) {
  buffer_put( buffer_2, s, strlen(s));
  buffer_put( buffer_2, "\n", 1);
  _exit(2);
}

void err_exit2(const char *s1,const char *s2) {
  buffer_put( buffer_2, s1, strlen(s1));
  err_exit(s2);
}

int open_read(const char *fname) {
  register int res;

  res = open(fname,O_RDONLY | O_NDELAY);
  if (res == -1) err_exit2("can't read-only open ",fname);
  return res;
}

int main(int argc,char **argv) {
  int linesfd;
  buffer srcbuf;
  char *srcname;
  char *linesname;
  off_t lineslen;
  char srcdatabuf[BUFFER_LEN+1];
  char *linesbuf;
  register unsigned long buf_lines;
  register const char **lines;

  if (argc<2) usage();
  linesname = argv[1];
  if (!linesname) usage();
  if (argc>2) {
    srcname = argv[2];
    if (!srcname) usage();
    buffer_init(&srcbuf, srcdatabuf, sizeof srcdatabuf, open_read(srcname));
  } else {
    buffer_init(&srcbuf, srcdatabuf, sizeof srcdatabuf, 0);
  }

  linesfd = open_read(linesname);
  lineslen = lseek(linesfd,0,SEEK_END);   /* EOF */
  if (lineslen == -1) err_exit2("can't lseek ",linesname);
  if (lseek(linesfd,0,SEEK_SET) == -1) err_exit2("can't lseek ",linesname);
  if (lineslen < 2) err_exit2("file too small: ",linesname);

  linesbuf = malloc(lineslen);
  if (!linesbuf) err_exit("malloc file failed");
  if (read(linesfd,linesbuf,lineslen) < 1) err_exit("can't read file into memory");

  buf_lines = line_count(linesbuf,lineslen);
  lines = malloc(buf_lines*sizeof(linesname));
  if (!lines) { free(linesbuf); err_exit("malloc array failed"); }
  line_array_fill(linesbuf,lineslen,lines);

  find_lines(lines,buf_lines,&srcbuf);
  buffer_flush(buffer_1);
  buffer_flush(buffer_2);
  if (srcbuf.fd!=0) close(srcbuf.fd);

  free(lines);
  free(linesbuf);
  _exit(0);
}
