/*
 * log dialin hosts IP# by calling an
 * internet connected HTTP-server cgi-script
 * and print the returned IP#
 *
 * compile with i.e.
 *   diet -Os gcc -Os -Wall -o loghostip loghostip.c
 *
 * Frank Bergmann, 3/2005
 */

#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/*
 * we call the server on IP# STR_SERVER_IP on tcp port 80
 * and send the following HTTP-command:
 *    GET /cgi-bin/logip?my-private-webserver-at-home HTTP/1.1
 *    Host: www.wasenmeister.de
 *
 * The IP# of our client my-private-webserver-at-home will be
 * logged on the server by the following CGI-script logip:
 *    #!/bin/sh
 *    echo "Content-Type: text/plain
 *    "
 *    echo $REMOTE_ADDR
 *    echo `date "+%Y%m%d %H%M%S"` $REMOTE_ADDR $QUERY_STRING >>iplog
 * The last logged IP# can be retrieved by commands like
 *    awk '$4==my-private-webserver-at-home {print$3}' iplog |tail -1
 *
 * After sending the HTTP command this program will read the
 * output from the HTTP-server (output of the script logip), strip
 * HTTP-header and bogus stuff and print the IP# of the client
 * it runs on to stdout.
 */
#define STR_SERVER_IP "129.0.0.99"
#define STR_HTTP_GET1 "GET /cgi-bin/logip?"
#define STR_HTTP_GET2 " HTTP/1.1\nHost: www.wasenmeister.de\n\n"

void fstrcat(register char *d,register char *s) {
  while(*d!='\0') d++;
  while(*s!='\0') *d++=*s++;
  *d='\0';
}

int len(register char *s) {
  register char *b;

  b=s;
  while(*b!='\0') b++;
  return b-s;
}

int main() {
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char hostname[12];
  char http_buf[513];
  char *p;

  if(inet_aton( STR_SERVER_IP, &server_sockaddr.sin_addr)==0) exit(1);
  server_sockaddr.sin_family= AF_INET;
  server_sockaddr.sin_port= htons( 80);
  if((sockfd=socket(PF_INET,SOCK_STREAM,0))==-1) exit(2);
  if(connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr))==-1) exit(3);
  gethostname(hostname,11);
  *http_buf='\0';
  fstrcat(http_buf,STR_HTTP_GET1);
  fstrcat(http_buf,hostname);
  fstrcat(http_buf,STR_HTTP_GET2);
  if(write(sockfd,http_buf,len(http_buf))==-1) exit(4);
  if(read(sockfd,http_buf,511)==-1) exit(5);
  /* print last *complete* line with at least 7 chars */
  p=http_buf+len(http_buf)-1;
  while((*p!='\n')&&(*(p-2)!='\n')) p--; p--;
  p[0]='\0';
  p-=7; while(*p!='\n') p--; p++;
  if(write(1,p,len(p))==-1) exit(6);
  if(close(sockfd)==-1) exit(7);
  exit(0);
}
