/*
 * logrockantennetitel-0.17.c
 * log titels of Radio Rockantenne
 *
 * compile with i.e.
 *   diet -Os gcc -Os -Wall -o logrockantennetitel logrockantennetitel.c; sstrip logrockantennetitel
 * best run with DJB's daemontools
 * script to create logrockantennetitel daemontools dir:

#!/bin/bash

mkdir rockantennelogd
mv logrockantennetitel rockantennelogd/
cd rockantennelogd
mkdir supervise
ln -s logrockantennetitel run
DIR=`pwd` 
cd /service
ln -s $DIR

 *
 * C programming style inspired by DJB and FeFe (i.e. libdjb, libowfat, fnord)
 * Frank Bergmann, 5/2008
 */

#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

/*
 * uncomment this for some debugging output to stderr
 *
 */
#define DEBUG 1

/*
 * uncomment this if you want ALL titles be logged (including i.e. "Nachrichten")
 *
 */
#define LOG_ALL_TITLES

/*
 * TIMEOUT in seconds for select socket-connect and socket-read
 * and alarm-timer for read()-kernel-hangups
 *
 */
#define TIMEOUT_SECS 30

/*
 * uncomment the following line if you want to use select()
 * note: we already use alarm() to catch also kernel-hangups on read()
 *
#define USE_SELECT
 */

/* #define STR_SERVER_IP "195.30.94.134" */
#define STR_SERVER_IP "194.97.153.233"
#define STR_HTTP_GET "GET /rock/datenbankspez/titel.txt HTTP/1.1\nHost: www.rockantenne.de\n\n"
#ifdef LOG_ALL_TITLES
#define INTERPRET_STR "&interpret="
#else
#define INTERPRET_STR "&interpret=>>"
#endif
#define TITLE_STR "&title="
#define NEXT_STR "&target="
#define BUFLEN 8192
#define STRING_LEN 128

/*
 * after TICKET_MAX_COUNT reset old title and log any new title,
 * don't care if it is the same title
 *
 * mostly this will be an error at rockantenne server, but maybe
 * it's our fault and therefore log any new title even if it is
 * the same title like 5 minutes before or 90 minutes before
 */
#define TICKET_MAX_COUNT 600

static jmp_buf procstate;

static void sig_alarm_handler(int sig) {
  longjmp(procstate,1);
}

int str_len(const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (int)(t-in);
}

int str_diffn(const char* a, const char* b, unsigned long limit) {
  register const unsigned char* s=(const unsigned char*)a;
  register const unsigned char* t=(const unsigned char*)b;
  register const unsigned char* u=t+limit;
  register int j;
  j=0;
  for (;;) {
    if (t>=u) break;
    if ((j=(*s-*t))) break;
    if (!*t) break;
    ++s; ++t;
  }
  return j;
}

int str_copyn(char *out,const char *in, unsigned long limit) {
  register char* s=out;
  register const char* t=in;
  register const char* u=t+limit;
  for (;;) {
    if (t>=u) { *s='\0'; break; }
    if (!(*s=*t)) break;
    ++s; ++t;
  }
  return (int)(s-out);
}

void str_catc(char* out, const char* in, const char sep) {
  register char* s=out;
  register const char* t=in;
  for (;;) { if (!*s) break; ++s; }
  for (;;) { if (!*t) break; *s++=*t++; }
  *s++=sep;
  *s='\0';
}

int str_index(const char* a, const char* b) {
  register const char* s=a;
  register int t=str_len(b);
  register const char* u=s+str_len(s)-t;
  for (;;) {
    if (s>u) break;
    if (!str_diffn(s,b,t)) return s-a;
    ++s;
  }
  return -1;
}

#ifdef DEBUG
static inline void write2cln(const char c) {
  char s[2];
  s[1]='\n';
  *s=c;
  write(2, s, 2);
}

static inline void write2s(const char *s) {
  write(2, s, str_len(s));
}
#endif

int do_http_get(char *buf, int buflen) {
  char get_string[]=STR_HTTP_GET;
  struct sockaddr_in server_sockaddr;
  volatile int sockfd;
  register int bl=buflen;
  register char *b=buf;
  register int i=0;
  register int newlines=0;
  int ret;
#ifdef USE_SELECT
  fd_set fds; 
  struct timeval tv; 
#endif

  sockfd=-1;
  if (setjmp(procstate)) { ret=-11; goto bad_get; }
  alarm(TIMEOUT_SECS);
  if (inet_aton( STR_SERVER_IP, &server_sockaddr.sin_addr)==0) { ret=-1; goto bad_get; }
  server_sockaddr.sin_family= AF_INET;
  server_sockaddr.sin_port= htons( 80);
  if ((sockfd=socket(PF_INET,SOCK_STREAM,0))==-1) { ret=-2; goto bad_get; }
#ifdef USE_SELECT
  fcntl(sockfd, F_SETFL,O_RDWR|O_NONBLOCK);
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr))<0) {
    if (errno != EINPROGRESS) { ret=-3; goto bad_get; }
  }
  tv.tv_sec = TIMEOUT_SECS;
  tv.tv_usec = 0;
  FD_ZERO(&fds);
  FD_SET(sockfd, &fds);
  if (select(sockfd+1, NULL, &fds, NULL, &tv) <= 0) { ret=-4; goto bad_get; }
  fcntl(sockfd, F_SETFL,O_RDWR);
#else
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr))<0) { ret=-5; goto bad_get; }
#endif
  if (write(sockfd,get_string,str_len(get_string))==-1) { ret=-6; goto bad_get; }
/*
  if ((bl=read(sockfd,b,buflen))<3) {
    if (errno != EAGAIN) { ret=-7; goto bad_get; }
  } else {
    goto data_read;
  }
  tv.tv_sec = TIMEOUT_SECS;
  tv.tv_usec = 0;
  FD_ZERO(&fds);
  FD_SET(sockfd, &fds);
  if (select(sockfd+1, &fds, NULL, NULL, &tv) <= 0) { ret=-9; goto bad_get; }
*/
  if ((bl=read(sockfd,b,buflen))<3) { ret=-10; goto bad_get; }
  /* fcntl(sockfd, F_SETFL,O_RDWR); */
  alarm(0);

/*
  data_read:
*/
  b[bl]='\0';
  for (;;) {
    if ((newlines>1) || (i>=bl)) break;
    if (b[i]!='\r') {
      if (b[i]=='\n') ++newlines; else newlines=0;
    }
    ++i;
  }
  if (newlines>1) {
    ret=i;
  }
  else {
#ifdef DEBUG
    write2s("no http header end\n");
#endif
    ret=-8;
  }

  bad_get:
  alarm(0);
  if (sockfd!=-1) close(sockfd);
  return ret; /* no end of http header found */
}

int get_song(char *http_buf, char **interpret, char **title) {
  register char *p;
  register int i;
  char ip_str[]=INTERPRET_STR;
  char tt_str[]=TITLE_STR;
  char nx_str[]=NEXT_STR;

  i = do_http_get(http_buf, BUFLEN);
  if (i<=0) {
#ifdef DEBUG
    write2s("error on do_http_get\n");
    write2cln('0'-i);
#endif
    return -1;
  }
  p = http_buf + i;
  i = str_index(p, ip_str);
  if (i<0) return -1;
  p += i + str_len(ip_str);
  while ((*p==' ') || (*p=='>')) ++p;
  *interpret = p;
  i = str_index(p, tt_str);
  if (i<0) return -1;
  p[i] = '\0';
  p += i + str_len(tt_str);
  while (*p==' ') ++p;
  *title = p;
  i = str_index(p, nx_str);
  if (i<0) return -1;
  p[i] = '\0';
  return 0;
}

static unsigned int fmt_2digits(char *s, int t) {
  s[0]=(t/10)+'0';
  s[1]=(t%10)+'0';
  return 2;
}

int get_time_string(char *s) {
  register char *p=s;
  time_t tp;
  struct tm *t;

  tp = time(NULL);
  if (tp<10) tp = 0;
  t = localtime(&tp);
  p += fmt_2digits(p, t->tm_year-100);
  p += fmt_2digits(p, t->tm_mon+1);
  p += fmt_2digits(p, t->tm_mday);
  p += fmt_2digits(p, t->tm_hour);
  p += fmt_2digits(p, t->tm_min);
  /* small delay correction */
  if (t->tm_sec>0) t->tm_sec -= 1;
  p += fmt_2digits(p, t->tm_sec);
  *p++ = '\t';
  *p = '\0';
#ifdef DEBUG
  if (t->tm_min==0) write2s("Ding-Dong!\n");
#endif
  if (t->tm_min>56) return (-1*t->tm_sec);
  return t->tm_sec;
}

int main() {
  register int i;
  register int sleep_secs;
  register int tickets;
  char *interpret, *title;
  char last_interpret[STRING_LEN+1], last_title[STRING_LEN+1];
  char last_interpret2[STRING_LEN+1], last_title2[STRING_LEN+1];
  char out_str[3*STRING_LEN];
  char http_buf[BUFLEN+1];
  int secs;

  signal(SIGALRM,sig_alarm_handler); /* fallback DFLT is OK */
  sleep_secs = 60;
  tickets = 0;
  *last_interpret = '\0'; *last_title = '\0';
  *last_interpret2 = '\0'; *last_title2 = '\0';
  for (;;) {
    interpret = last_interpret;
    title = last_title;
    secs = get_time_string(out_str);
    if (secs<0) sleep_secs = 59 + secs;
    i = get_song(http_buf, &interpret, &title);
    if (i!=0) {
#ifdef DEBUG
      write2s("error on get_song\n");
#endif
      goto do_sleep;
    }
    if ( (str_diffn(interpret, last_interpret, STRING_LEN)==0) &&
        (str_diffn(title, last_title, STRING_LEN)==0) ) {
      goto do_sleep;
    }
    if ( (str_diffn(interpret, last_interpret2, STRING_LEN)==0) &&
        (str_diffn(title, last_title2, STRING_LEN)==0) ) {
#ifdef DEBUG
      write2s("before last equal\n");
#endif
      goto do_sleep;
    }
    if ((secs<0) && (str_diffn(interpret, "Nachrichten", 11)==0)) {
#ifdef DEBUG
      write2s("secs<0, Nachrichten\n");
#endif
      goto do_sleep;
    }
    tickets = 0;
    str_catc(out_str, interpret, '\t');
    str_catc(out_str, title, '\n');
#ifdef DEBUG
    write(2, out_str, str_len(out_str));
#endif
    i = open("/var/log/rockantenne.log", O_WRONLY | O_APPEND | O_CREAT,
        S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
    if (i>0) {
      write(i, out_str, str_len(out_str));
      close(i);
    }
    str_copyn(last_interpret2,last_interpret,STRING_LEN);
    str_copyn(last_title2,last_title,STRING_LEN);
    str_copyn(last_interpret,interpret,STRING_LEN);
    str_copyn(last_title,title,STRING_LEN);
    /* if (sleep_secs==SMALL_SLEEP) sleep_secs = 109; */
    /* rockantenne seems to use a minutely cronjob, wait til next minute starts */
    sleep_secs = 59 - secs;
    if (str_diffn(interpret, "Nachrichten", 11)==0) sleep_secs += 120;

  do_sleep:
    if (secs>6) sleep_secs = 59 - secs;
    sleep(sleep_secs);
    sleep_secs = 1;
    tickets++;
    if (tickets>TICKET_MAX_COUNT) {
#ifdef DEBUG
      write2s("ticketcount reachs maximum\n");
#endif
      tickets = 0;
      *last_interpret = '\0'; *last_title = '\0';
      *last_interpret2 = '\0'; *last_title2 = '\0';
    }
  }
}
