/*
 * showip.c
 * log dialin hosts IP# by calling an
 * internet connected HTTP-server cgi-script
 * and print the returned IP#
 * (used as dyndns service on tuxad.net)
 *
 * Used for:
 * - getting my own dynamic or NAT IP address with a simple
 *   http-get, most compatible with firewalls
 * (- log it (IP# & name) for remote access or for statistics)
 * (- register my name at tuxad.net dyndns service)
 *
 * compile with i.e.
 *   diet -Os gcc -Os -Wall -o showip showip.c;strip showip
 *
 * Frank Bergmann, 1/2006
 */

#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/*
 * we call the server on IP# STR_SERVER_IP on tcp port 80
 * and send the following HTTP-command:
 *    GET /showip HTTP/1.1
 *    Host: dyndns.tuxad.net
 * The IP# of our client my-private-webserver-at-home will be
 * created on the server by the following CGI-script logip:
 *    #!/bin/sh
 *    echo "Content-Type: text/plain
 *    "
 *    echo $REMOTE_ADDR
 * After sending the HTTP command this program will read the
 * output from the HTTP-server (output of the script logip), strip
 * HTTP-header and bogus stuff and print the IP# of the client
 * it runs on to stdout.
 */
#define STR_SERVER_IP "81.89.239.232"
#define STR_HTTP_GET "GET /showip HTTP/1.1\nHost: dns.tdyn.de\n\n"
#define READBYTES 255

int len(const char* in) {
  register const char* t=in;
  for (;;) { if (!*t) break; ++t; }
  return (int)(t-in);
}

int main() {
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char http_buf[2*READBYTES];
  char get_string[]=STR_HTTP_GET;
  register char *p;
  register ssize_t buflen;

  if(inet_aton( STR_SERVER_IP, &server_sockaddr.sin_addr)==0) exit(1);
  server_sockaddr.sin_family= AF_INET;
  server_sockaddr.sin_port= htons( 80);
  if((sockfd=socket(PF_INET,SOCK_STREAM,0))==-1) exit(2);
  if(connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr))==-1) exit(3);
  if(write(sockfd,get_string,len(get_string))==-1) exit(4);
  p=http_buf;
  if((buflen=read(sockfd,p,READBYTES))==-1) exit(5);
  p+=buflen;
  /* maybe we got only the header, try another read */
  buflen=read(sockfd,p,READBYTES);
  if(buflen>0) p+=buflen;
  close(sockfd);
  for(;;p--) if((*p<32)&&(*(p-1)>32)) break;
  *p=0;
  for(;;) {
    p--; if (*p<32) break;
  }
  p++;
  write(1,p,len(p));
  exit(0);
}
