/*
 * showrockantennetitel-0.4.c
 * show title currently played on Radio Rockantenne
 *
 * compile with i.e.
 *   diet -Os gcc -Os -Wall -o showrockantennetitel showrockantennetitel.c; strip showrockantennetitel
 *
 * Frank Bergmann, 3/2006
 */

#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>

/*
 * uncomment this if you want a TIMEOUT in seconds for socket-connect
 *
 */
#define TIMEOUT_SECS 10

/* #define STR_SERVER_IP "195.30.94.134" */
#define STR_SERVER_IP "194.97.153.233"
#define STR_HTTP_GET "GET /rock/datenbankspez/titel.txt HTTP/1.1\nHost: www.rockantenne.de\n\n"
#define INTERPRET_STR "&interpret="
#define TITLE_STR "&title="
#define NEXT_STR "&target="
#define BUFLEN 8192
#define STRING_LEN 128

struct song_data {
  char *interpret;
  char *title;
};

int str_len(const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (int)(t-in);
}

int str_diffn(const char* a, const char* b, unsigned long limit) {
  register const unsigned char* s=(const unsigned char*)a;
  register const unsigned char* t=(const unsigned char*)b;
  register const unsigned char* u=t+limit;
  register int j;
  j=0;
  for (;;) { if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t; }
  return j;
}

int str_index(const char* a, const char* b) {
  register const char* s=a;
  register int t=str_len(b);
  register const char* u=s+str_len(s)-t;
  for (;;) {
    if (s>u) break;
    if (!str_diffn(s,b,t)) return s-a;
    ++s;
  }
  return -1;
}

#ifdef DEBUG
static inline void write2cln(const char c) {
  char s[2];
  s[1]='\n';
  *s=c;
  write(2, s, 2);
}

static inline void write2s(const char *s) {
  write(2, s, str_len(s));
}
#endif

int do_http_get(char *buf, int buflen) {
  char get_string[]=STR_HTTP_GET;
  struct sockaddr_in server_sockaddr;
  int sockfd;
  register int bl=buflen;
  register char *b=buf;
  register int i=0;
  register int newlines=0;
#ifdef TIMEOUT_SECS
  int ret;
  fd_set wfds; 
  struct timeval tv; 
#endif

  sockfd=-1;
  if (inet_aton( STR_SERVER_IP, &server_sockaddr.sin_addr)==0) { ret=-1; goto badget; }
  server_sockaddr.sin_family= AF_INET;
  server_sockaddr.sin_port= htons( 80);
  if ((sockfd=socket(PF_INET,SOCK_STREAM,0))==-1) {
#ifdef DEBUG
    write2s("rockantenne: error creating socket\n");
#endif
    ret=-2; goto badget;
  }
  fcntl(sockfd, F_SETFL,O_RDWR|O_NONBLOCK);
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr))<0) {
    if (errno != EINPROGRESS) {
#ifdef DEBUG
      write2s("rockantenne: error on connect\n");
#endif
      ret=-3; goto badget;
    }
  }
  tv.tv_sec = TIMEOUT_SECS;
  tv.tv_usec = 0;
  FD_ZERO(&wfds);
  FD_SET(sockfd, &wfds);
  if (select(sockfd+1, NULL, &wfds, NULL, &tv) <= 0) { ret=-4; goto badget; }
  fcntl(sockfd, F_SETFL,O_RDWR);
  if (write(sockfd,get_string,str_len(get_string))==-1) { ret=-6; goto badget; }
  if ((bl=read(sockfd,b,buflen))<3) { ret=-7; goto badget; }
  b[bl]='\0';
  for (;;) {
    if ((newlines>1) || (i>=bl)) break;
    if (b[i]!='\r') {
      if (b[i]=='\n') ++newlines; else newlines=0;
    }
    ++i;
  }
  if (newlines>1) {
    ret=i;
  }
  else {
#ifdef DEBUG
    write2s("rockantenne: no http header end\n");
#endif
    ret=-8;
  }

  badget:
  if (sockfd!=-1) close(sockfd);
  return ret; /* no end of http header found */
}

int get_song_data(char *http_buf, int index, struct song_data *song) {
  register int i;
  register char *p;
  char ip_str[]=INTERPRET_STR;
  char tt_str[]=TITLE_STR;
  char nx_str[]=NEXT_STR;

  i = index;
  p = http_buf + i;
  i = str_index(p, ip_str);
  if (i<0) return 10;
  p += i + str_len(ip_str);
  while ((*p==' ') || (*p=='>')) ++p;
  song->interpret = p;
  i = str_index(p, tt_str);
  if (i<0) return 10;
  p[i] = ':';
  p[i+1] = '\0';
  p += i + str_len(tt_str);
  while (*p==' ') ++p;
  song->title = p;
  i = str_index(p, nx_str);
  if (i<0) return 10;
  p[i++] = '\n';
  p[i] = '\0';
  return 0;
}

int main() {
  int i;
  struct song_data song;
  char http_buf[BUFLEN+1];

  i = do_http_get(http_buf, BUFLEN);
  if (i>0) {
    i = get_song_data(http_buf, i, &song);
    if (i) return i;
    write(1, song.interpret, str_len(song.interpret));
    write(1, song.title, str_len(song.title));
    return 0;
  }
  return 1;
}
