/* 
   sstrip: Copyright (C) 1999-2001 by Brian Raiter, under the GNU
   General Public License. No warranty. See COPYING for details.

   some changes by Nikola Vladov - Copyright (C) 2007
   disable: printf, malloc, strerror, write to tmpfile.
   uses only system_functions + memcpy + strlen
   diet -Os gcc -s -o sstrip sstrip.c -Wall -W
 */

/* define next for files larger than 2 G. */  
/*#define _FILE_OFFSET_BITS 64*/

#include	<unistd.h>
#include	<alloca.h>
#include	<stdlib.h>
#include	<string.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<elf.h>
#include	<asm/elf.h>
#include	<sys/stat.h>

#ifndef TRUE
#define	TRUE		1
#define	FALSE		0
#endif

#if ELF_CLASS == ELFCLASS32
#define	Elf_Ehdr	Elf32_Ehdr
#define	Elf_Phdr	Elf32_Phdr
#else
#define	Elf_Ehdr	Elf64_Ehdr
#define	Elf_Phdr	Elf64_Phdr
#endif

/* The name of the program. */
static char *progname;

/* The name of the current file. */
static char *filename;
static char flagdebug;

#ifdef __linux__
#include	<sys/sendfile.h>
#else
#include	<sys/mman.h>
off_t sendfile(int out, int in, off_t *offset, size_t bytes) {
  char* map;
  int w;
  off_t mofs, done=0;
  while (bytes) {
    mofs=*offset&~0xffff;
    map=mmap(0, 0x10000, PROT_READ, MAP_SHARED, in, mofs);
    if (map==MAP_FAILED) return -1;
    w=write(out, map + (*offset&0xffff), (bytes>0x10000) ? 0x10000: bytes);
    munmap(map, 0x10000);
    if (w<=0) break;
    bytes-=w;
    *offset+=w;
    done+=w;
  }
  return done;
}
#endif

static char *X_strerror(int e) {
  static char buf[20];
  unsigned long u;
  char *x;
  switch (e) {
  case EPERM:   x = "Operation not permitted"; break;
  case ENOENT:	x = "No such file"; break;
  case EACCES:	x = "Permission denied"; break;
  case EISDIR:	x = "Is a directory"; break;
  case ENOMEM:	x = "Out of memory"; break;
  case ENOSPC:	x = "No space left on device"; break;
  case ELOOP:	x = "Too many symbolic links"; break;
  case EINVAL:	x = "Invalid argument"; break;
  case EIO:	x = "I/O error"; break;
  default:
    x = buf+19; *x = 0; u = e;
    do { --x; *x = '0' + (u%10); u /= 10; } while (u);
    x -= 6;
    memcpy(x, "error=", 6); 
  }
  return x;
}

/* Last pointer must be zero !!! */
static void X_print(int fd, char *s, ...) {
  char **argv = &s;
  for (; (s=*argv); argv++) write(fd, s, strlen(s));
}

#define MAX_STATIC_FU_NMB	4
char *fu(unsigned long u) {
  static char fu_buffer[MAX_STATIC_FU_NMB * 12];
  static char k=MAX_STATIC_FU_NMB;
  char *p = fu_buffer + 12*k;
  *--p = 0;
  do { *--p = '0' + (u%10); u /= 10; } while (u);
  if (--k == 1) k=MAX_STATIC_FU_NMB;
  return p;
}

static void format_header(Elf_Ehdr *ehdr, Elf_Phdr *phdrs) {
  Elf_Phdr *phdr = phdrs;
  char x[8];
  size_t i;
  X_print(1, filename, ":\n\tsegmets:\toffset=", fu(ehdr->e_phoff),
	  "\tsize=", fu(ehdr->e_phnum * sizeof *phdr), "\n", 0);
  for (i=0; i < ehdr->e_phnum; i++, phdr++) {
    memcpy(x, "\t---\n", 6);
    if (phdr->p_flags & PF_R) x[1] = 'r';
    if (phdr->p_flags & PF_W) x[2] = 'w';
    if (phdr->p_flags & PF_X) x[3] = 'x';
    X_print(1, "\tsegment: ", fu(i), 
	    "\toffset=", fu(phdr->p_offset),
	    "\tsize=", fu(phdr->p_filesz), x, 0);
  }
}

/* A simple error-handling function. FALSE is always returned for the
 * convenience of the caller.
 */
static int err(char const *errmsg) {
  X_print(2, progname, ": ", filename, ": ", errmsg, "\n", 0);
  return FALSE;
}

/* The given error message is used only when errno is not set.
 */
int ferr(char *msg) { return err(errno ? X_strerror(errno) : msg); }

/* readelfheader() reads the ELF header into our global variable, and
 * checks to make sure that this is in fact a file that we should be
 * munging.
 */
static int readelfheader(int fd, Elf_Ehdr *ehdr) {
    if (read(fd, ehdr, sizeof *ehdr) != sizeof *ehdr)
	return ferr("missing or incomplete ELF header");

    /* Check the ELF signature.
     */
    if (!(ehdr->e_ident[EI_MAG0] == ELFMAG0 &&
	  ehdr->e_ident[EI_MAG1] == ELFMAG1 &&
	  ehdr->e_ident[EI_MAG2] == ELFMAG2 &&
	  ehdr->e_ident[EI_MAG3] == ELFMAG3))
	return err("missing ELF signature");

    /* Compare the file's class and endianness with the program's.
     */
    if (ehdr->e_ident[EI_DATA] != ELF_DATA)
	return err("ELF file has different endianness");
    if (ehdr->e_ident[EI_CLASS] != ELF_CLASS)
	return err("ELF file has different word size");

    /* Check the target architecture/version.
     */
    if (ehdr->e_machine != ELF_ARCH || ehdr->e_version != EV_CURRENT)
	return err("ELF file created for different architecture/version");

    /* Verify the sizes of the ELF header and the program segment
     * header table entries.
     */
    if (ehdr->e_ehsize != sizeof(Elf_Ehdr))
	return err("unrecognized ELF header size");
    if (ehdr->e_phentsize != sizeof(Elf_Phdr))
	return err("unrecognized program segment header size");

    /* Finally, check the file type.
     */
    if (ehdr->e_type != ET_EXEC && ehdr->e_type != ET_DYN)
	return err("not an executable or shared-object library");

    if (!ehdr->e_phnum || ehdr->e_phoff < sizeof *ehdr)
	return ferr("ELF file has no program header table");
    
    return TRUE;
}

/* getmemorysize() determines the offset of the last byte of the file
 * that is referenced by an entry in the program segment header table.
 * (Anything in the file after that point is not used when the program
 * is executing, and thus can be safely discarded.)
 */
static int getmemorysize(int fd, Elf_Ehdr *ehdr, Elf_Phdr *phdrs,
			 size_t *newsize) {
    Elf_Phdr const   *phdr;
    size_t		size, n, i;

    if (sizeof *ehdr < ehdr->e_phoff &&
	lseek(fd, ehdr->e_phoff, SEEK_SET) == (off_t)-1) return ferr("");
    n = ehdr->e_phnum * sizeof *phdrs;
    if (read(fd, phdrs, n) != (ssize_t)n) 
        return ferr("missing or incomplete program segment header table");

    if (flagdebug) format_header(ehdr, phdrs);
    /* Start by setting the size to include the ELF header and the
     * complete program segment header table.
     */
    size = ehdr->e_phoff + n;

    /* Then keep extending the size to include whatever data the
     * program segment header table references.
     */
    for (i = 0, phdr = phdrs ; i < ehdr->e_phnum ; ++i, ++phdr) {
	if (phdr->p_type != PT_NULL && phdr->p_filesz) {
	    n = phdr->p_offset + phdr->p_filesz;
	    if (n > size)
		size = n;
	}
    }

    *newsize = size;
    return TRUE;
}

/* truncatezeros() examines the bytes at the end of the file's
 * size-to-be, and reduces the size to exclude any trailing zero
 * bytes.
 */
static int truncatezeros(int fd, size_t *newsize) {
    unsigned char	contents[1024];
    size_t		size, n;

    size = *newsize;
    do {
      n = sizeof contents;
      if (n > size) n = size;
      if (lseek(fd, size - n, SEEK_SET) == (off_t)-1) return ferr("");
      if (read(fd, contents, n) != (ssize_t)n) return ferr("short read");
      while (n && !contents[--n]) --size;
    } while (size && !n);

    *newsize = size;
    return TRUE;
}

/* modifyheaders() removes references to the section header table if
 * it was stripped, and reduces program header table entries that
 * included truncated bytes at the end of the file.
 */
static int modifyheaders(Elf_Ehdr *ehdr, Elf_Phdr *phdrs, size_t newsize) {
    Elf_Phdr *phdr;
    size_t	i;

    /* Sanity check. */
    if (newsize < ehdr->e_phoff + ehdr->e_phnum * sizeof *phdrs)
        return err("ELF file is completely blank");

    /* If the section header table is gone, then remove all references
     * to it in the ELF header.
     */
    if (ehdr->e_shoff >= newsize) {
	ehdr->e_shoff = 0;
	ehdr->e_shnum = 0;
	/* ehdr->e_shentsize = 0; */
	/* fbergmann: */
	/* to avoid "corrupted section header size" set it to 40 */
	ehdr->e_shentsize = 40;
	ehdr->e_shstrndx = 0;
    }

    /* The program adjusts the file size of any segment that was
     * truncated. The case of a segment being completely stripped out
     * is handled separately.
     */
    for (i = 0, phdr = phdrs ; i < ehdr->e_phnum ; ++i, ++phdr) {
	if (phdr->p_offset >= newsize) {
	    phdr->p_filesz = 0;
	} else if (phdr->p_offset + phdr->p_filesz > newsize) {
	    phdr->p_filesz = newsize - phdr->p_offset;
	}
    }
    return TRUE;
}

/* commitchanges() writes the new headers back to the original file
 * and sets the file to its new size.
 */
static int commitchanges(int fd, Elf_Ehdr *ehdr, Elf_Phdr *phdrs,
			 size_t newsize) {
    size_t	n,k;
    int		fdout=-1;
    char       *tmpname, *x, *q;
    off_t	offset = 0;
    struct stat st;

    if (fstat(fd, &st)) return ferr("");
    if (st.st_size <= newsize) return TRUE;

    if (flagdebug) {
      X_print(1, "\told=",fu(st.st_size), 
	      "\tnew=",fu(newsize), 
	      "\tdiff=",fu(st.st_size -newsize), "\n",0);
      format_header(ehdr, phdrs);
      if (flagdebug > 1) return TRUE;
    }

    n = strlen(filename);
    tmpname = alloca(n + 32);
    if (tmpname==0) { errno=ENOMEM; return ferr(""); }
    memcpy(tmpname, filename, n);
    x = tmpname + n;
	 
    for (n=0; ; n++) {
      for (k=n, q=x; k; ) { *q++ = 'A' + (k%16); k /= 16; }
      q[0] = '~'; q[1] = 0;
      fdout = open(tmpname, O_WRONLY|O_CREAT|O_EXCL, st.st_mode);
      if (fdout >= 0 || errno != EEXIST) break;
    }
    if (fdout == -1) return err("unable to open tmpfile");
    x = "";
    if (sendfile(fdout, fd, &offset, newsize) != (off_t)newsize)
      { x = "sendfile: short write"; goto warning; }

    /* Save the changes to the ELF header, if any. */
    if (lseek(fdout, 0, SEEK_SET) == (off_t)-1) goto warning;
    if (write(fdout, ehdr, sizeof *ehdr) != sizeof *ehdr)
      { x = "short write"; goto warning; }

    /* Save the changes to the program segment header table, if any. */
    n = ehdr->e_phnum * sizeof *phdrs;
    if (sizeof *ehdr < ehdr->e_phoff &&
	lseek(fdout, ehdr->e_phoff, SEEK_SET) == (off_t)-1) goto warning;
    if (write(fdout, phdrs, n) != (ssize_t)n)
      { x = "short write"; goto warning; }

    /* We opened fdout in the right mode already */
    if (fchown(fdout, st.st_uid, st.st_gid)) goto warning;
    fsync(fdout);
    if (close(fdout)) goto warning;
    if (rename(tmpname, filename)) goto warning;

    return TRUE;
  warning:
    ferr(x);
    if (fdout>=0) { close(fdout); unlink(tmpname); }
    return FALSE;
}

/* Here we use alloca() for phdrs each time */
static int X_main() {
  int			fd;
  Elf_Ehdr		ehdr;
  Elf_Phdr	       *phdrs;
  size_t		newsize=0;

  fd = open(filename, O_RDONLY);
  if (fd < 0) { ferr("can't open"); return 1; }
  if (!(readelfheader(fd, &ehdr))) goto ops;

  phdrs = alloca(ehdr.e_phnum * sizeof *phdrs);
  if (!phdrs) { errno=ENOMEM; ferr(""); goto ops; }

  if ((getmemorysize(fd, &ehdr, phdrs, &newsize)	&&
       truncatezeros(fd, &newsize)			&&
       modifyheaders(&ehdr, phdrs, newsize)		&&
       commitchanges(fd, &ehdr, phdrs, newsize))) {
    close(fd);
    return 0;
  } 

 ops:
  close(fd);
  return 1;
}

int main(int argc, char **argv) {
  char	      **arg;
  int		failures = 0;

  arg = argv+1;  
  if (argc < 2) {
  usage:
    X_print(1, "Usage: [-s|d] sstrip FILE...\n"
      "\tsstrip discards all nonessential bytes from an executable.\n\n"
      "\tVersion 2.0 Copyright (C) 2000,2001 Brian Raiter.\n"
      "\tHacked for dietlibc, Copyright (C) 2007 Nikola Vladov.\n\n"
      "\tThis program is free software, licensed under the GNU\n"
      "\tGeneral Public License.  There is absolutely no warranty.\n", 0);
    return EXIT_SUCCESS;
  }

  if (arg[0][0] == '-') {
    switch (arg[0][1]) {
    case 'd': flagdebug = 1; break;
    case 's': flagdebug = 2; break;
    default: goto usage;
    }
    ++arg;
  }
  progname = argv[0];

  for (; *arg != NULL ; ++arg) {
    filename = *arg;
    failures |= X_main(filename);
  }
  return failures;
}
