/*
 * utail-0.2.c
 * a very minimalistic tail
 * its primary goal is to show the last 4 lines of rockantenne.log,
 * served through DJBs tcpserver and controlled with daemontools,
 * rockantenne.log can be written by logrockantennetitel
 *
 * compile with i.e.
 *   diet -Os gcc -Os -Wall -o utail utail.c; strip utail
 *
 * best run with DJB's daemontools
 * script to create nettailrockantenne daemontools dir and start service:

#!/bin/bash
mkdir -p nettailrockantenne/supervise nettailrockantenne/root
mv utail nettailrockantenne/root
cd nettailrockantenne
cat >run <<EOF
#!/bin/sh
exec 2>&1
cd root
exec /usr/local/bin/tcpserver -v -RHl localhost -u nobody -g nobody 0 4488 ./utail -4 /var/log/rockantenne.log
EOF
chmod 755 run
DIR=`pwd` 
cd /service
ln -s $DIR
# the data can be retrieved through the net by i.e. "nc localhost 4488"
# <script end>

 *
 * C programming style inspired by DJB and FeFe (i.e. libdjb, libowfat, fnord)
 *
 * Frank Bergmann, 6/2006
 */

#define _FILE_OFFSET_BITS 64
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

#define BUFLEN 8192
#define INLEN 32

const char filename_error[] = "error: no filename given\n";
const char open_error[] = "error: can't open file\n";
const char seek_error[] = "error: can't seek file\n";
const char read_error[] = "error: can't read file\n";

int str_len(const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (int)(t-in);
}

void usage() {
  char u[] = "utail [-n N] FILE\n";
  write(2, u, str_len(u));
  exit(127);
}

int do_utail(char *fname, unsigned long lines) {
  register char *p;
  register int nlcount;
  ssize_t bytes;
  char buf[BUFLEN+1];
  int fd;
  off_t fpos;

  fd = open(fname, O_RDONLY);
  if (fd == -1) {
    write(2, open_error, str_len(open_error));
    return 2;
  }
  fpos = lseek(fd, 0, SEEK_END);
  if (fpos == -1) {
    write(2, seek_error, str_len(seek_error));
    return 3;
  }
  fpos -= BUFLEN;
  if (fpos < 0) fpos = 0;
  if (lseek(fd, fpos, SEEK_SET) == -1) {
    write(2, seek_error, str_len(seek_error));
    return 3;
  }
  p = buf;
  bytes = read(fd, p, BUFLEN);
  close(fd);
  if (bytes < 0) {
    write(2, read_error, str_len(read_error));
    return 4;
  }
  p += (bytes-2); /* -2: skip EOF and last nl -- last line has maybe no nl */
  if (p < buf) return 0; /* no line data */
  nlcount = 0;
  for(;;) {
    if(*p=='\n')nlcount++;if(nlcount==lines)break;p--;if(p<buf)break;
    if(*p=='\n')nlcount++;if(nlcount==lines)break;p--;if(p<buf)break;
    if(*p=='\n')nlcount++;if(nlcount==lines)break;p--;if(p<buf)break;
    if(*p=='\n')nlcount++;if(nlcount==lines)break;p--;if(p<buf)break;
  }
  p++;
  write(1, p, (buf+bytes)-p);
  return 0;
}

unsigned long n2a(const char*c, int exit_on_error) {
  unsigned long l=0;
  while (isdigit(*c)) {
    l=l*10+*c-'0';
    ++c;
  }
  if (*c && exit_on_error) usage();
  return l;
}

int main(int argc, char *argv[]) {
  int i,j;
  char inp[INLEN+1];
  unsigned long lines;

  lines = 10;
  for(i=1;i<argc;++i) {
    if (argv[i][0]=='-') {
      switch (argv[i][1]) {
      case '0' ... '9':
	lines=n2a(&argv[i][1], 1);
	break;
      case 'n':
        if (argv[i][2]=='-') { /* -n- reads 'lines' from stdin */
          j = read(0, inp, INLEN);
          inp[INLEN] = '\0';
          if (j>0) {
            lines = n2a(inp, 0);
            if (lines < 1) lines = 1;
          }
        } else {
	  if (argv[i][2]) lines=n2a(&argv[i][2],1); else lines=n2a(argv[++i],1);
        }
	break;
      default:
	usage();
	break;
      }
    }
    else break;
  }
  if ((i+1)!=argc) {
    usage();
  }
  return do_utail(argv[i], lines);
}
